"use strict";

const ee = require("./eventemitter");
const eventNamespace = "tgam-photo-slider";
const reqTimeout = require("./request-timeout");
const lazyConfig = require("./lazysizes-config");

const throttle = require("./throttle.js");
throttle("resize", "windowResize.throttled", window);

// component photo slider
const slideHasInfoClass = "c-photo-slider--has-info";
const slideHasNavClass = "c-photo-slider--has-nav";
const slideHasShareToolsClass = "c-photo-slider--has-share-tools";

const slideProgressBarClass = "c-photo-slider__progress-bar";
const slideProgressTextClass = "c-photo-slider__progress-text";

const slideListClass = "c-photo-slider__list";
const slideItemClass = "c-photo-slider__item";

const slideNavLeftClass = "c-photo-slider__nav--left";
const slideNavRightClass = "c-photo-slider__nav--right";

const slideNavButtonLeftClass = "c-photo-slider__nav-button--left";
const slideNavButtonRightClass = "c-photo-slider__nav-button--right";

const slideShareToolsClass = "c-photo-slider__share-tools";

// component article media image
const slideMediaClass = "c-article-media__image";
const slideMediaLinkClass = "c-article-media__link";
const slideMediaLinkLoadedClass = "c-article-media__link--loaded";
const slideMediaInfoClass = "c-article-media__info";
const slideMediaCenterClass = "c-article-media__center";

// component image
const slideImageClass = "c-image";
const invisibleClass = "u-invisible";

module.exports = {

  init: function(rootElement, rootClass, slideStart = 0) {
    if (!rootElement) { return; }

    this.rootElement = rootElement;
    this.rootClass = rootClass; // js hook class name
    this.slideActive = slideStart; // current slide shown
    this.sliding = false; // block overlaping slide animations

    this.slideSpeed = 200; // ms: speed when animating between slides
    this.slideMargin = 20; // px: margin between slides

    this.landscapeMax = 1020; // px: only applied to landscape photos
    this.portraitMax = 940; // px: only applied to portrait photos
    this.baseScale = 100; // px: default width of photos, then scaled up or down

    this.enableSwipeDismiss = true; // true if swipe (up or down) dismiss allowed?
    this.lockAxis = 0; // enum: x-axis = 1, y-axis = -1, neither = 0

    this.origX = 0; // px: original position of touch start event x/y axis
    this.origY = 0;

    this.distX = 0; // px: distance travelled of touch move x/y axis
    this.distY = 0;

    this.velocityThreshold = 0.5; // swipe velocity threshold for axis - for flick swipes
    this.distanceThreshold = 200; // px: swipe distance threshold for axis - for slow swipes

    // unique element id for this slider (that will be prefixed with various string)
    this.uniqueId = (0|Math.random()*9e6).toString(36); // e.g. "1p7ym"

    this.slideProgressBar = rootElement.getElementsByClassName(slideProgressBarClass).item(0);
    this.slideProgressText = rootElement.getElementsByClassName(slideProgressTextClass).item(0);
    
    this.slideList = rootElement.getElementsByClassName(slideListClass).item(0);
    this.slideShareTools = rootElement.getElementsByClassName(slideShareToolsClass).item(0);
    this.slideItems = this.arrayFrom(rootElement.getElementsByClassName(slideItemClass));
    this.slides = [];

    if (!this.slideList || !this.slideItems) { return; }

    this.slideItems.forEach(item => {
      const slideType = [...item.classList].reduce((accumulator, className) => {
        if (className.includes("c-photo-slider__item--")) {
          return className.split("--")[1];
        }
      });
      this.slides.push({
        root: item,
        type: slideType,
        image: item.getElementsByClassName(slideImageClass).item(0),         // formerly slideItemImages
        imageLink: item.getElementsByClassName(slideMediaLinkClass).item(0), // formerly slideItemLinks
        info: item.getElementsByClassName(slideMediaInfoClass).item(0),      // formerly slideItemInfos
        media: item.getElementsByClassName(slideMediaCenterClass).item(0)    // formerly slideItemMedias
      });
    });

    // target of the left/right button aria-controls
    this.slideList.id = `c-photo-slider-${this.uniqueId}`;

    rootElement.addEventListener("lazybeforesizes", this.onLazyBeforeSizes.bind(this));

    // this will cause placeholder image to be positioned
    this.processPlaceholder(this.slideActive);

    // trigger this manually to get intial measurements
    this.onSliderWindowResize();

    // go to default slide
    this.slideToTouchEvent(this.slideActive);

    // grab all images, we don't want any ghosting for any of them
    this.slideItemGhosts = this.arrayFrom(rootElement.getElementsByTagName("img"));
    window.requestAnimationFrame(this.onHideGhostImages.bind(this));

    // setup aria attributes for links
    this.slides.forEach((slide, index) => {
      if (slide.imageLink) {
        const id = this.slideList.id + "-" + (index + 1);
        slide.imageLink.setAttribute("aria-role", "button");
        slide.imageLink.setAttribute("aria-label", "Toggle section");
        // for related information overlay
        if (slide.info) {
          slide.imageLink.setAttribute("aria-expanded", "true");
          slide.imageLink.setAttribute("aria-controls", id);
          slide.info.id = id;
        }
      }
    });
    
    // all touch event bindings
    this.boundOnSliderTouchStart = this.onSliderTouchStart.bind(this);
    this.boundOnSliderTouchMove = this.onSliderTouchMove.bind(this);
    this.boundOnSliderTouchEnd = this.onSliderTouchEnd.bind(this);
    
    // delegate the slide element click event to root element
    this.boundOnSliderClick = this.onSliderClick.bind(this);
    
    // keypress events for navigation
    this.boundOnRootElementKeyDown = this.onRootElementKeyDown.bind(this);

    // previous slide button
    this.slideNavLeft = rootElement.getElementsByClassName(slideNavLeftClass).item(0);
    this.slideNavButtonLeft = rootElement.getElementsByClassName(slideNavButtonLeftClass).item(0);
    if (this.slideNavLeft && this.slideNavButtonLeft) {
      this.slideNavButtonLeft.setAttribute("aria-controls", this.slideList.id);
      this.boundOnSlideNavLeft = this.onSlideNavLeft.bind(this);
    }

    // next slide button
    this.slideNavRight = rootElement.getElementsByClassName(slideNavRightClass).item(0);
    this.slideNavButtonRight = rootElement.getElementsByClassName(slideNavButtonRightClass).item(0);
    if (this.slideNavRight && this.slideNavButtonRight) {
      this.slideNavButtonRight.setAttribute("aria-controls", this.slideList.id);
      this.boundOnSlideNavRight = this.onSlideNavRight.bind(this);
    }
    
    // setup initial left/right button states
    window.requestAnimationFrame(this.updateButtonStates.bind(this));

    // keep slider container width updated
    window.addEventListener("windowResize.throttled", this.onSliderWindowResize.bind(this));

    this.unlockSlides();
  },

  // placeholder "src" image is displayed before lazysizes loads an image
  processPlaceholder: function(index) {
    const imageIndex = this.safeIndex(this.slides, index);

    if (this.slides[imageIndex].image) {
      const placeImage = this.slides[imageIndex].image;
      const newWidth = this.calculateSize(placeImage, placeImage.width);
      const newScale = this.calculateScale(newWidth);

      window.requestAnimationFrame(() => {
        placeImage.style.width = this.baseScale + "px"; // retain this value
        placeImage.style.transform = `scale(${newScale})`;
      });
    }
  },

  onLazyBeforeSizes: function(e) {
    if (e.defaultPrevented) {
      return;
    }
    // find the slide element and object that are associated with this image
    const slideElement = e.target.closest("." + slideItemClass);
    if (!slideElement) {
      return;
    }
    let slideIndex = null;
    this.slides.forEach((slide, index) => {
      if (slide.root === slideElement) {
        slideIndex = index;
      }
    });
    if (slideIndex !== null && this.slides[slideIndex] && this.slides[slideIndex].image) {
      const image = this.slides[slideIndex].image;
      const media = this.slides[slideIndex].media;
      const newWidth = this.calculateSize(image, e.detail.width);
      const newScale = this.calculateScale(newWidth);

      // must set this to bubble up into lazysizes
      e.detail.width = newWidth;

      window.requestAnimationFrame(() => {
        // must update element style width for chrome render bug
        // note: "sizes" attribute is updated within lazysizes (values should match)
        image.style.transform = `scale(${newScale})`;

        // update figure caption container to match image width
        if (media) {
          media.style.width = newWidth + "px"; // can't use transform for this element :(
          media.classList.remove(invisibleClass); // caption is hidden until image is loaded
        }
      });
    }
  },

  // update image style scale to match detail width, which also matches sizes attribute
  calculateScale: function(width) {
    return !width ? 1 : (width / this.baseScale);
  },

  // update image style width to match detail width, which also matches sizes attribute
  calculateSize: function(element, imageWidth) {
    if (!element) { return imageWidth; } //  portrait [_] < 100 > [___] landscape
    const imageRatio = parseFloat(element.dataset.aspectRatio || 1);
    const fitElem = element.parentNode || element; // container should be immediate parent
    let maxHeight = fitElem.clientHeight;
    let maxWidth = fitElem.clientWidth;

    // maximum values depending on image orientation
    if (imageRatio > 100) {
      maxWidth = Math.min(maxWidth, this.landscapeMax); // for landscape images
    } else {
      maxHeight = Math.min(maxHeight, this.portraitMax); // for portrait images
    }

    let ret = maxWidth; // return new width value
    const displayRatio = (maxWidth / maxHeight) * 100; // same formula as back-end

    // basically if display is wider than image, and
    // the image height is no longer contained within display height
    if (displayRatio > imageRatio) {
      ret = (imageRatio / displayRatio) * maxWidth;
    }

    return Math.round(ret);
  },

  // toggle the image figure caption section
  toggleInterface: function() {
    window.requestAnimationFrame(function onSliderToggleInterfaceAnimation() {
      if (this.hasInfoShown()) {
        this.hideInfo();
      } else {
        this.showInfo();
      }
    }.bind(this));
  },

  hasInfoShown: function() {
    return this.rootElement.classList.contains(slideHasInfoClass);
  },

  hasNavShown: function() {
    return this.rootElement.classList.contains(slideHasNavClass);
  },

  showNav: function() {
    // only show navigation if more than one slide present
    if (this.slides.length) {
      this.rootElement.classList.add(slideHasNavClass);
    }
  },

  hideNav: function() {
    this.rootElement.classList.remove(slideHasNavClass);
  },

  showInfo: function() {
    this.rootElement.classList.add(slideHasInfoClass);
    this.slides.map(slide => {
      if (slide.imageLink) {
        slide.imageLink.setAttribute("aria-expanded", "true");
      }
      if (slide.info) {
        slide.info.removeAttribute("aria-hidden");
      }
    });
    ee.emitEvent(`${eventNamespace}:toggleinfo`, [this]);
  },

  hideInfo: function() {
    this.rootElement.classList.remove(slideHasInfoClass);
    this.slides.map(slide => {
      if (slide.imageLink) {
        slide.imageLink.setAttribute("aria-expanded", "false");
      }
      if (slide.info) {
        slide.info.setAttribute("aria-hidden", "true");
      }
    });
    ee.emitEvent(`${eventNamespace}:toggleinfo`, [this]);
  },

  hideShareTools: function() {
    this.rootElement.classList.remove(slideHasShareToolsClass);
    this.slideShareTools.classList.add(invisibleClass);
  },

  updateButtonStates: function() {
    if (this.slideActive === 0) {
      this.slideNavButtonLeft.setAttribute("aria-disabled", "true");
      this.slideNavButtonLeft.setAttribute("disabled", "disabled");
    } else {
      this.slideNavButtonLeft.removeAttribute("aria-disabled");
      this.slideNavButtonLeft.removeAttribute("disabled");
    }
    if (this.slideActive === this.slides.length - 1) {
      this.slideNavButtonRight.setAttribute("aria-disabled", "true");
      this.slideNavButtonRight.setAttribute("disabled", "disabled");
    } else {
      this.slideNavButtonRight.removeAttribute("aria-disabled");
      this.slideNavButtonRight.removeAttribute("disabled");
    }
  },
  
  unlockSlides: function() {
    // all touch event bindings
    this.rootElement.addEventListener("touchstart", this.boundOnSliderTouchStart);
    this.rootElement.addEventListener("touchmove", this.boundOnSliderTouchMove);
    this.rootElement.addEventListener("touchend", this.boundOnSliderTouchEnd);
    
    // delegate the slide element click event to root element
    this.slideList.addEventListener("click", this.boundOnSliderClick);
    
    // keypress events for navigation
    this.rootElement.addEventListener("keydown", this.boundOnRootElementKeyDown);
    
    if (this.slideNavLeft && this.slideNavButtonLeft) {
      this.slideNavLeft.addEventListener("keydown", this.boundOnSlideNavLeft);
      this.slideNavLeft.addEventListener("click", this.boundOnSlideNavLeft);
    }
    if (this.slideNavRight && this.slideNavButtonRight) {
      this.slideNavRight.addEventListener("keydown", this.boundOnSlideNavRight);
      this.slideNavRight.addEventListener("click", this.boundOnSlideNavRight);
    }
    // enable buttons
    this.updateButtonStates();
  },

  lockSlides: function(timeout = null, unlockCallback = null) {
    if (timeout) {
      window.setTimeout(() => {
        this.unlockSlides();
        if (unlockCallback) {
          unlockCallback();
        }
      }, timeout);
    }
    // all touch event bindings
    this.rootElement.removeEventListener("touchstart", this.boundOnSliderTouchStart);
    this.rootElement.removeEventListener("touchmove", this.boundOnSliderTouchMove);
    this.rootElement.removeEventListener("touchend", this.boundOnSliderTouchEnd);
    
    // delegate the slide element click event to root element
    this.slideList.removeEventListener("click", this.boundOnSliderClick);
    
    // keypress events for navigation
    this.rootElement.removeEventListener("keydown", this.boundOnRootElementKeyDown);
    
    if (this.slideNavLeft && this.slideNavButtonLeft) {
      this.slideNavLeft.removeEventListener("keydown", this.boundOnSlideNavLeft);
      this.slideNavLeft.removeEventListener("click", this.boundOnSlideNavLeft);
    }
    if (this.slideNavRight && this.slideNavButtonRight) {
      this.slideNavRight.removeEventListener("keydown", this.boundOnSlideNavRight);
      this.slideNavRight.removeEventListener("click", this.boundOnSlideNavRight);
    }
    // disable buttons
    this.slideNavButtonLeft.setAttribute("aria-disabled", "true");
    this.slideNavButtonLeft.setAttribute("disabled", "disabled");
    this.slideNavButtonRight.setAttribute("aria-disabled", "true");
    this.slideNavButtonRight.setAttribute("disabled", "disabled");
    
    // start timer animation
    this.startTimer(timeout);
  },

  startTimer: function(timeout) {
    let start = null;
    const rootElement = this.rootElement;
    const progressBar = this.slideProgressBar;
    const progressText = this.slideProgressText;
    
    progressText.innerHTML = "0%";
    progressBar.style.width = 0;
    progressBar.setAttribute("aria-valuenow", 0);
    progressBar.removeAttribute("hidden");
    rootElement.setAttribute("aria-busy", "true");
    rootElement.setAttribute("aria-describedby", progressBar.id);
    
    function step(timestamp) {
      if (!start) {
        start = timestamp;
      }
      const progress = timestamp - start;
      const percentFinished = (Math.round(progress) / timeout) * 100;
      
      progressText.innerHTML = Math.round(percentFinished) + "%";
      progressBar.style.width = percentFinished + "%";
      progressBar.setAttribute("aria-valuenow", Math.round(percentFinished));
      
      if (progress < timeout) {
        window.requestAnimationFrame(step);
      } else {
        progressBar.setAttribute("hidden", "hidden");
        rootElement.removeAttribute("aria-busy");
        rootElement.removeAttribute("aria-describedby");
      }
    }
    window.requestAnimationFrame(step);
  },

  isKeySpaceOrEnter: function(e) {
    let match = false;
    switch(e.keyCode) {
      case 32: // spacebar
      case 13: // enter
        if (!e.metaKey) {
          return true;
        }
        break;
    }
    return false;
  },

  // simple polyfill for Array.from if neccessary
  arrayFrom: function(HTMLCollection) {
    if (!Array.from) {
      return (HTMLCollection ? [].slice.call(HTMLCollection) : []);
    } else {
      return Array.from(HTMLCollection);
    }
  },

  onSliderWindowResize: function() {
    // transition is slowest for touch input
    this.slideWidth = document.body.clientWidth + this.slideMargin;
    // transition is faster for mouse input
    this.clickWidth = this.slideWidth / 10;
    // transition is fastest (instant) for keyboard input
    this.keyWidth = 0;
    // default to slide width when setting up other slide positions
    this.setupSlides(this.slideActive, this.slideWidth);
  },

  onSliderTouchStart: function(e) {
    this.startTimeTouch = new Date().getTime();
    this.origX = e.touches[0].pageX;
    this.origY = e.touches[0].pageY;
    this.distX = 0;
    this.distY = 0;
    this.setupSlides(this.slideActive, this.slideWidth);
  },

  onSliderTouchMove: function(e) {
    e.preventDefault(); // block browser zoom and scroll (to allow below swiping)

    this.distX = e.touches[0].pageX - this.origX;
    this.distY = e.touches[0].pageY - this.origY;

    let absDistX = Math.abs(this.distX);
    let absDistY = Math.abs(this.distY);

    // minimum threshold for axis movement
    this.distX = (absDistX > (this.distanceThreshold / 8)) ? this.distX : 0;
    this.distY = (absDistY > (this.distanceThreshold / 2)) ? this.distY : 0;

    this.lockAxis = 0; // no movement
    if (absDistX > absDistY || !this.enableSwipeDismiss) {
      this.lockAxis = 1; // x-axis (use intends to change slide)
    } else if (absDistY > absDistX) {
      this.lockAxis = -1; // y-axis (use intends to dismiss)
    }

    if (this.lockAxis !== 0) {
      window.requestAnimationFrame(function onSliderTouchMoveAnimation() {
        this.translateSlide(this.slideActive, this.lockAxis, this.distX, this.distY, this.slideWidth);
      }.bind(this));
    }
  },

  onSliderTouchEnd: function(e) {
    let absDistX = Math.abs(this.distX);
    let absDistY = Math.abs(this.distY);
    let timespan = (new Date().getTime() - this.startTimeTouch);
    let speedX = absDistX / timespan; // touchmove speed in px/ms
    let speedY = absDistY / timespan; // touchmove speed in px/ms

    if (absDistX > absDistY) {
      if (this.distX > 0) {
        if (this.slideActive > 0 &&
          (absDistX > this.distanceThreshold || speedX > this.velocityThreshold)) {
          this.slideToTouchEvent(this.slideActive - 1); // right
        } else {
          this.slideToTouchEvent(this.slideActive); // cancel
        }
      } else if (this.distX < 0) {
        if (this.slideActive < this.slides.length - 1 &&
          (absDistX > this.distanceThreshold || speedX > this.velocityThreshold)) {
          this.slideToTouchEvent(this.slideActive + 1); // left
        } else {
          this.slideToTouchEvent(this.slideActive); // cancel
        }
      }
    } else if (this.enableSwipeDismiss) {
      if (this.distY < 0) {
        if (absDistY > (this.distanceThreshold * 1.5) || speedY > this.velocityThreshold) {
          this.closeAnimation(this.slideActive, this.distX, true); // up
        } else {
          this.slideToTouchEvent(this.slideActive); // cancel
        }
      } else if (this.distY > 0) {
        if (absDistY > (this.distanceThreshold * 1.5) || speedY > this.velocityThreshold) {
          this.closeAnimation(this.slideActive, this.distX, false); // down
        } else {
          this.slideToTouchEvent(this.slideActive); // cancel
        }
      }
    }
  },

  // animate the photo slider to appear that it swipes up or down (dismissed)
  closeAnimation: function(slideActive, offsetX, upward) {
    window.requestAnimationFrame(function onCloseAnimationStart() {
      // hide the ui components
      this.hideShareTools();
      this.hideInfo();
      this.hideNav();
      // perform the translation up or down
      let positionY = (upward ? -1 : 1) * (this.rootElement.clientHeight);
      this.translateSlide(slideActive, -1, offsetX, positionY, this.slideWidth);
    }.bind(this));

    // fire event after above tranlsate is complete
    reqTimeout.setTimeout(function closeAnimationComplete() {
      ee.emitEvent(`${eventNamespace}:swipeup`, [this]);
    }.bind(this), this.slideSpeed / 2);
  },

  // triggered by touch screen input
  slideToTouchEvent: function(index) {
    this.slideTo(index, false, this.slideWidth);
  },

  // triggered by mouse input
  slideToClickEvent: function(index) {
    this.slideTo(index, false, this.clickWidth);
  },

  // triggered by keyboard input
  slideToKeyEvent: function(index) {
    this.slideTo(index, true, this.keyWidth);
  },

  // slide index is zero-based
  // for certain trigger events, you would not want to set focus on the new slide
  // for certain trigger events, animation is instant
  slideTo: function(slideIndex, setFocus, slideWidth) {
    if (this.sliding) { return; }
    this.sliding = true;

    // validate requested slide index
    this.slideActive = this.safeIndex(this.slides, slideIndex);

    // setup the slide animation move
    window.requestAnimationFrame(function onSlideToAnimation() {
      this.translateSlide(this.slideActive, 0, 0, 0, slideWidth);
      this.updateSlideStates();
      this.updateButtonStates();
      this.preloadSlide(this.slideActive + 1);
      this.preloadSlide(this.slideActive - 1);

      // a setTimeout is needed here, since we want to set focus below, after the slide is fully in view.
      // otherwise the focus would cause the viewport to jump ahead of the animation.
      reqTimeout.setTimeout(function onSlideToTimeout() {
        // only set focus when event did not originate from mouse click
        // so set focus for touch input and key input
        if (setFocus && this.slides[this.slideActive].imageLink) {
          this.slides[this.slideActive].imageLink.focus();
        }
        ee.emitEvent(`${eventNamespace}:selected`, [this]);
        this.setupSlides(this.slideActive, slideWidth);
        this.sliding = false;
      }.bind(this), this.slideSpeed);
    }.bind(this));
  },

  // setup the initial positions of the neighbouring slides (excluding the current slide)
  setupSlides: function(slideIndex, slideWidth) {
    let prevIndex = slideIndex - 1;
    let nextIndex = slideIndex + 1;

    let nextX = slideWidth;
    let prevX = (-1 * slideWidth);

    this.slides.forEach((slide, index) => {
      if (index > this.slideActive) {
        slide.root.style.transform = `translate(${nextX}px,0px)`;
      } else
      if (index < this.slideActive) {
        slide.root.style.transform = `translate(${prevX}px,0px)`;
      }
    });
  },

  // incrementally position the previous/current/next slide(s)
  translateSlide: function(slideIndex, lockAxis, offsetX, offsetY, slideWidth) {
    offsetX = (lockAxis === -1) ? 0 : offsetX; // y-axis locked?
    offsetY = (lockAxis === 1) ? 0 : offsetY; // x-axis locked?

    let prevIndex = slideIndex - 1;
    let nextIndex = slideIndex + 1;

    let nextX = slideWidth + offsetX;
    let prevX = (-1 * slideWidth) + offsetX;

    if (this.slides[slideIndex]) {
      this.slides[slideIndex].root.style.transform = `translate(${offsetX}px,${offsetY}px)`;
    }
    if (this.slides[nextIndex]) {
      this.slides[nextIndex].root.style.opacity = 1;
      this.slides[nextIndex].root.style.transform = `translate(${nextX}px,${offsetY}px)`;
    }
    if (this.slides[prevIndex]) {
      this.slides[prevIndex].root.style.opacity = 1;
      this.slides[prevIndex].root.style.transform = `translate(${prevX}px,${offsetY}px)`;
    }
  },

  updateSlideStates: function() {
    this.slides.forEach((slide, index) => {
      if (index == this.slideActive) {
        slide.root.style.opacity = 1;
        slide.root.removeAttribute("aria-hidden");
        if (slide.imageLink) {
          slide.imageLink.removeAttribute("tabindex");
        }
      } else {
        slide.root.style.opacity = 0;
        slide.root.setAttribute("aria-hidden", "true");
        if (slide.imageLink) {
          slide.imageLink.setAttribute("tabindex", "-1");
        }
      }
    });
  },

  onSlideNavLeft: function(e) {
    if (this.hasNavShown()) {
      if (e.type === "click" || (e.type === "keydown" && this.isKeySpaceOrEnter(e))) {
        e.preventDefault();
        e.stopPropagation();
        this.slideToClickEvent(this.slideActive - 1); // previous
      }
    }
  },

  onSlideNavRight: function(e) {
    if (this.hasNavShown()) {
      if (e.type === "click" || (e.type === "keydown" && this.isKeySpaceOrEnter(e))) {
        e.preventDefault();
        e.stopPropagation();
        this.slideToClickEvent(this.slideActive + 1); // next
      }
    }
  },

  onHideGhostImages: function() {
    // prevent all images from showing ghosts when dragged via mouse
    this.slideItemGhosts.map(element => {
      element.setAttribute("draggable", "false");
    });
  },

  onSliderClick: function(e) {
    e.preventDefault();
    this.toggleInterface();
  },

  onRootElementKeyDown: function(e) {
    let match = false;
    switch(e.keyCode) {
      case 34: // page down
      case 39: // right
      case 40: // down
        match = true;
        this.slideToKeyEvent(this.slideActive + 1); // next
        break;

      case 33: // page up
      case 37: // left
      case 38: // up
        match = true;
        this.slideToKeyEvent(this.slideActive - 1); // previous
        break;

      case 36: // home
        match = true;
        this.slideToKeyEvent(0); // first
        break;

      case 35: // end
        match = true;
        this.slideToKeyEvent(this.slides.length - 1); // last
        break;

      // delegate the slide element events to root element
      case 32: // spacebar
      case 13: // enter
        match = true;
        this.toggleInterface();
        break;
    }
    // allow all other key events bubble
    if (match && !e.metaKey) {
      e.preventDefault();
      e.stopPropagation();
    }
  },

  // public method
  // generate HTML for a photo slider component
  generatePhotoSlider: function(galleryPhotos, params) {
    if (!galleryPhotos) {
      return;
    }

    // default options
    const defaults = {
      jsHookClass: "js-c-photo-slider",
      svgSpritePath: "sprite.svg",
      shareToolsClone: true,
      shareToolsClass: "c-share-tools",
      shareToolsItemClass: "c-share-tool",
      shareToolsFilterClass: "js-c-photo-slider-remove"
    };
    const options = this.merge(defaults, params);

    let slideList = ""; // all slide list items
    let index = 0; // index of current slide item

    for (let i = 0; i < galleryPhotos.length; i++) {
      if (galleryPhotos[i].slideType && galleryPhotos[i].slideType === "ad") {
        slideList += `
          <li class="c-photo-slider__item" data-slide-type="ad">
            <div class="c-photo-slider__item-content">
              ${galleryPhotos[i].slideMarkup}
            </div>
          </li>`;
      } else {
        let slideImage = this.generateImage(index, galleryPhotos[i].placeholderSrc, galleryPhotos[i].src, galleryPhotos[i].srcset, galleryPhotos[i].aspectRatio, galleryPhotos[i].alt)
        let slideCaption = this.generateCaption(index, galleryPhotos.length, galleryPhotos[i].caption, galleryPhotos[i].credit);
        slideList += this.generateListItem(slideImage, slideCaption);
      }
      index++;
    }

    // get first share tools from [article] page
    // if none found, then photo slider will not have share tools
    let toolsClone = null;
    if (options.shareToolsClone) {
      let shareToolsSource = document.getElementsByClassName(options.shareToolsClass).item(0);
      if (shareToolsSource) {
        toolsClone = this.generateShareTools(shareToolsSource, options.shareToolsItemClass, options.shareToolsFilterClass);
      }
    }

    return this.generateSlideDeck(options.jsHookClass, options.svgSpritePath, slideList, toolsClone);
  },

  // generate HTML for a single slide image
  generateImage: function(index, placeholderSrc, dataSrc, dataSrcSet, aspectRatio, altText) {
    // base64 encode of 1x1px transparent GIF
    const hasSrc = placeholderSrc ? `src="${placeholderSrc}"` : `src="data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7"`;
    const hasDataSrc = dataSrc ? `data-src="${dataSrc}"` : "";
    const hasDataSrcSet = dataSrcSet ? `data-srcset="${dataSrcSet}"` : "";
    const hasAspectRatio = `data-aspect-ratio="${aspectRatio}"`
    const hasAlt = altText ? `alt="${altText}"` : "";
    return `
      <a href="#" class="c-article-media__link">
        <img data-index="${index}"
          class="${slideImageClass} ${lazyConfig.lazyClass}"
          ${hasSrc}
          ${hasDataSrc}
          data-sizes="auto"
          ${hasDataSrcSet}
          ${hasAspectRatio}
          ${hasAlt} />
        <div class="c-spinner c-spinner-lg c-spinner-light"></div>
      </a>`;
  },

  // generate HTML for a single image figure caption
  generateCaption: function(index, total, caption = "", credit = "") {
    const hasMediaCaption = caption ? `<p class="c-article-media__caption">${caption}</p>` : "";
    const hasMediaCredit = credit ? `<p class="c-article-media__credit">${credit}</p>` : "";
    return `
      <figcaption class="c-article-media__info c-article-media__info--dark c-article-media__info--center">
        <div class="c-article-media__center u-invisible">
          <p class="c-article-media__label">${index + 1} of ${total}</p>
          <div class="c-article-media__description">
            ${hasMediaCaption}
            ${hasMediaCredit}
          </div>
        </div>
      </figcaption>`;
  },

  // generate the HTML for a single slide list item
  generateListItem: function(slideImage = "", slideCaption = "") {
    return `
      <li class="c-photo-slider__item" data-slide-type="photo">
        <div class="c-photo-slider__item-content">
          <figure class="c-article-media c-article-media--overlay">
            <div class="c-article-media__image c-article-media__image--full-fit">
              ${slideImage}
            </div>
            ${slideCaption}
          </figure>
        </div>
      </li>`;
  },

  // generate the HTML for the share tools section
  // extract the share tools from the article page
  // optionally remove items that contain filterClass
  generateShareTools: function(shareToolsElement, toolClass, filterClass) {
    if (!shareToolsElement) {
      return;
    }

    let shareItems = this.arrayFrom(shareToolsElement.getElementsByClassName(toolClass));
    if (filterClass) {
      shareItems = shareItems.filter((item) => {
        return !item.classList.contains(filterClass);
      });
    }
    shareItems = shareItems.map((item) => {
      return item.outerHTML;
    });

    return `
      <div class="c-photo-slider__share-tools">
        <div class="c-share-tools c-share-tools--dark u-roofless u-floorless">
          ${shareItems.join("")}
        </div>
      </div>`;
  },

  // generate slide deck with: prev/next buttons, slides, and share tools
  generateSlideDeck: function(jsHookClass, svgSpritePath, slidesHtml, shareTools) {
    const hookClass = jsHookClass || "js-c-photo-slider";
    const hasShareTools = shareTools ? "c-photo-slider--has-share-tools" : "";
    return `
      <div class="c-photo-slider ${hookClass} ${hasShareTools} c-photo-slider--has-info c-photo-slider--has-nav"
        role="region" aria-label="Photo slideshow" tabindex="0">
        <div class="c-photo-slider__progress"></div>
        ${this.generateSlideButton(false, svgSpritePath)}
        ${this.generateSlideButton(true, svgSpritePath)}
        <ul class="c-photo-slider__list" aria-live="polite" aria-atomic="true">
          ${slidesHtml}
        </ul>
        ${shareTools}
      </div>`;
  },

  // generate a prev/next button
  generateSlideButton: function(next, svgSpritePath) {
    const direction = next ? "right" : "left";
    const label = next ? "next" : "previous";
    return `
      <div class="c-photo-slider__nav c-photo-slider__nav--${direction}">
        <button class="c-photo-slider__nav-button c-photo-slider__nav-button--${direction}" aria-label="View ${label} photo">
          <svg class="c-photo-slider__nav-icon" role="img" pointer-events="none" focusable="false" aria-hidden="true">
            <use xlink:href="${svgSpritePath}#icon-${direction}-chevron"></use>
          </svg>
        </button>
      </div>`;
  },

  // simple shallow extend/assign method
  merge: function(obj, src) {
    for (var key in src) {
      if (src.hasOwnProperty(key)) {
        obj[key] = src[key];
      }
    }
    return obj;
  },

  // public method
  // return unique identifier class name
  getRootClass: function() {
    return this.rootClass;
  },

  // public method
  // return current zero-based slide index
  getIndex: function() {
    return this.slideActive;
  },

  // public method
  // return current slide type
  getSlideType: function() {
    return this.slides[this.slideActive].type;
  },

  // add lazysizes preload class to neighbouring slides
  preloadSlide: function(number) {
    number = this.safeIndex(this.slides, number);
    this.processPlaceholder(number);

    if (this.slides[number].image) {
      this.slides[number].image.classList.add(lazyConfig.preloadClass);
    }
  },

  // return a safe index value for an array
  // within zero and the array length
  safeIndex: function(list, index) {
    if (!list || !Array.isArray(list) || index === null || typeof index === "undefined") {
      return 0;
    } else if (index < 0) {
      return 0;
    } else if (index >= list.length) {
      return list.length - 1;
    } else {
      return index;
    }
  }

};
